# -*- coding: utf-8 -*-
from __future__ import print_function

from gi.repository import GObject, Gtk, Gio

from dfeet.uiloader import UILoader
from dfeet.introspection import AddressInfo


class DBusBusName(GObject.GObject):
    """class to represent a name on the bus"""
    def __init__(self, bus_name_unique):
        super(DBusBusName, self).__init__()
        self.__bus_name_unique = bus_name_unique
        self.__pid = 0
        self.__cmdline = ''

    def __repr__(self):
        return "%s (pid: %s)" % (self.bus_name_unique, self.pid)

    def __update_cmdline(self):
        if self.pid > 0:
            procpath = '/proc/' + str(self.pid) + '/cmdline'
            with open(procpath, 'r') as f:
                self.__cmdline = " ".join(f.readline().split('\0'))
        else:
            self.__cmdline = ''

    @property
    def bus_name_unique(self):
        return self.__bus_name_unique

    @property
    def pid(self):
        return self.__pid

    @pid.setter
    def pid(self, pid_new):
        self.__pid = pid_new
        try:
            self.__update_cmdline()
        except:
            self.__cmdline = ''

    @property
    def cmdline(self):
        return self.__cmdline


class BusWatch(object):
    """watch for a given bus"""
    def __init__(self, data_dir, address):
        self.data_dir = data_dir
        self.address = address
        #setup UI
        ui = UILoader(self.data_dir, UILoader.UI_BUSWATCH)
        self.paned_buswatch = ui.get_root_widget()
        self.liststore_model = ui.get_widget('liststore_buswatch')
        self.treemodelfilter_buswatch = ui.get_widget('treemodelfilter_buswatch')
        self.treemodelfilter_buswatch.set_visible_func(self.__treemodelfilter_buswatch_cb)
        self.treemodelsort_buswatch = ui.get_widget("treemodelsort_buswatch")
        self.treemodelsort_buswatch.set_sort_func(2, self.__sort_on_name)
        self.treemodelsort_buswatch.set_sort_column_id(2, Gtk.SortType.ASCENDING)
        self.treeview = ui.get_widget('treeview_buswatch')
        self.entry_filter = ui.get_widget('entry_filter')
        self.grid_bus_name_selected_info = ui.get_widget('grid_bus_name_info')
        self.label_bus_name_selected_name = ui.get_widget('label_bus_name_selected_name')
        self.label_bus_name_selected_pid = ui.get_widget('label_bus_name_selected_pid')
        self.label_bus_name_selected_cmdline = ui.get_widget('label_bus_name_selected_cmdline')
        self.addr_info = None  # hold the currently selected AddressInfo object

        self.treeview.connect('cursor-changed',
                              self.__tree_view_cursor_changed_cb)
        self.entry_filter.connect("changed",
                                  self.__entry_filter_changed_cb)

        #setup the conection
        if self.address == Gio.BusType.SYSTEM or self.address == Gio.BusType.SESSION:
            self.connection = Gio.bus_get_sync(self.address, None)
        elif Gio.dbus_is_supported_address(self.address):
            self.connection = Gio.DBusConnection.new_for_address_sync(
                self.address,
                Gio.DBusConnectionFlags.AUTHENTICATION_CLIENT |
                Gio.DBusConnectionFlags.MESSAGE_BUS_CONNECTION,
                None, None)

        #setup signals
        self.connection.signal_subscribe(None, "org.freedesktop.DBus", "NameOwnerChanged",
                                         None, None, 0, self.__name_owner_changed_cb, None)

        self.bus_proxy = Gio.DBusProxy.new_sync(self.connection,
                                                Gio.DBusProxyFlags.NONE,
                                                None,
                                                'org.freedesktop.DBus',
                                                '/org/freedesktop/DBus',
                                                'org.freedesktop.DBus', None)

        #list all names
        self.bus_proxy.ListNames('()',
                                 result_handler=self.__list_names_handler,
                                 error_handler=self.__list_names_error_handler)

    def __treemodelfilter_buswatch_cb(self, model, iter, user_data):
        #return model.get_value(iter, 1) in data
        bus_name_obj = model.get(iter, 0)[0]
        filter_text = self.entry_filter.get_text()
        return filter_text.lower() in bus_name_obj.bus_name_unique.lower()

    def __entry_filter_changed_cb(self, entry_filter):
        self.treemodelfilter_buswatch.refilter()

    def __tree_view_cursor_changed_cb(self, treeview):
        """do something when a row is selected"""
        selection = self.treeview.get_selection()
        if selection:
            model, iter_ = selection.get_selected()
            if not iter_:
                return

            bus_name_obj = model.get(iter_, 0)[0]
            #remove current child
            c2 = self.paned_buswatch.get_child2()
            if c2:
                c2.destroy()
            try:
                del(self.addr_info)
            except:
                pass

            #add Introspection to paned
            self.addr_info = AddressInfo(
                self.data_dir, self.address, bus_name_obj.bus_name_unique, connection_is_bus=True)
            self.paned_buswatch.add2(self.addr_info.introspect_box)

            #update info about selected bus name
            self.label_bus_name_selected_name.set_text(bus_name_obj.bus_name_unique)
            self.label_bus_name_selected_pid.set_text("%s" % bus_name_obj.pid)
            self.label_bus_name_selected_cmdline.set_text(bus_name_obj.cmdline)
            self.grid_bus_name_selected_info.set_visible(True)

    def __liststore_model_add(self, bus_name_obj):
        """add a DBusBusName object to the liststore model"""
        #update bus info stuff
        self.bus_proxy.GetConnectionUnixProcessID(
            '(s)', bus_name_obj.bus_name_unique,
            result_handler=self.__get_unix_process_id_cb,
            error_handler=self.__get_unix_process_id_error_cb,
            user_data=bus_name_obj)
        #add bus to liststore
        return self.liststore_model.append(
            [bus_name_obj, 0, "%s" % (bus_name_obj.bus_name_unique), bus_name_obj.cmdline])

    def __liststore_model_remove(self, bus_name_obj):
        """remove a DBusBusName object to the liststore model"""
        for n, obj in enumerate(self.liststore_model):
            if obj[2] == bus_name_obj.bus_name_unique:
                del(self.liststore_model[n])
                break

    def __liststore_model_get(self, bus_name_obj):
        """get a object from the liststore"""
        for n, obj in enumerate(self.liststore_model):
            if obj[2] == bus_name_obj.bus_name_unique:
                return obj
        raise Exception("bus name object '%s' not found in liststore" % (bus_name_obj))

    def __name_owner_changed_cb(self, connection, sender_name,
                                object_path, interface_name, signal_name,
                                parameters, user_data):
        """bus name added or removed"""
        bus_name = parameters[0]
        old_owner = parameters[1]
        new_owner = parameters[2]

        bus_name_obj = DBusBusName(bus_name)

        if bus_name[0] == ':':
            if not old_owner:
                self.__liststore_model_add(bus_name_obj)
            else:
                self.__liststore_model_remove(bus_name_obj)
        else:
            if new_owner:
                self.__liststore_model_add(bus_name_obj)
            if old_owner:
                self.__liststore_model_remove(bus_name_obj)

    def __list_names_handler(self, obj, names, userdata):
        for n in names:
            bus_name_obj = DBusBusName(n)
            self.__liststore_model_add(bus_name_obj)

    def __list_names_error_handler(self, obj, error, userdata):
        print("error getting bus names: %s" % str(error))

    def __get_unix_process_id_cb(self, obj, pid, bus_name_obj):
        bus_name_obj.pid = pid

    def __get_unix_process_id_error_cb(self, obj, error, bus_name_obj):
        print(
            "error getting unix process id for %s: %s" % (
                bus_name_obj.bus_name_unique, str(error)))
        bus_name_obj.pid = 0

    def __sort_on_name(self, model, iter1, iter2, user_data):
        un1 = model.get_value(iter1, 2)
        un2 = model.get_value(iter2, 2)

        # covert to integers if comparing two unique names
        if un1[0] == ':' and un2[0] == ':':
            un1 = un1[1:].split('.')
            un1 = tuple(map(int, un1))

            un2 = un2[1:].split('.')
            un2 = tuple(map(int, un2))

        elif un1[0] == ':' and un2[0] != ':':
            return 1
        elif un1[0] != ':' and un2[0] == ':':
            return -1
        else:
            un1 = un1.split('.')
            un2 = un2.split('.')

        if un1 == un2:
            return 0
        elif un1 > un2:
            return 1
        else:
            return -1


if __name__ == "__main__":
    """for debugging"""
    import sys
    import argparse

    parser = argparse.ArgumentParser(description='show a given bus address')
    parser.add_argument('addr')
    p = parser.parse_args()

    if p.addr.lower() == 'system':
        addr = Gio.BusType.SYSTEM
    elif p.addr.lower() == 'session':
        addr = Gio.BusType.SESSION
    else:
        addr = p.addr

    bw = BusWatch(addr)

    win = Gtk.Window()
    win.connect("delete-event", Gtk.main_quit)
    win.set_default_size(1024, 768)
    win.add(bw.paned_buswatch)
    win.show_all()
    try:
        Gtk.main()
    except (KeyboardInterrupt, SystemExit):
        Gtk.main_quit()



# -*- coding: utf-8 -*-
# Copyright (C) 2013 Thomas Bechtold <thomasbechtold@jpberlin.de>

# This file is part of D-Feet.

# D-Feet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# D-Feet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with D-Feet.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject, Gtk, Gio

import gettext
from gettext import gettext as _
gettext.textdomain('d-feet')

from dfeet.bus_watch import BusWatch
from dfeet.settings import Settings
from dfeet.uiloader import UILoader
from dfeet.addconnectiondialog import AddConnectionDialog
from dfeet.executemethoddialog import ExecuteMethodDialog


class NotebookTabLabel(Gtk.Box):
    __gsignals__ = {
        "close-clicked": (GObject.SIGNAL_RUN_FIRST, GObject.TYPE_NONE, ()),
    }

    def __init__(self, label_text):
        Gtk.Box.__init__(self)
        self.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.set_spacing(5)
        # label
        label = Gtk.Label(label_text)
        self.pack_start(label, True, True, 0)
        # close button
        button = Gtk.Button()
        button.set_relief(Gtk.ReliefStyle.NONE)
        button.set_focus_on_click(False)
        button.add(Gtk.Image.new_from_stock(Gtk.STOCK_CLOSE, Gtk.IconSize.MENU))
        button.connect("clicked", self.__button_clicked)
        self.pack_end(button, False, False, 0)
        self.show_all()

    def __button_clicked(self, button, data=None):
        self.emit("close-clicked")


class DFeetWindow(Gtk.ApplicationWindow):
    """the main window"""

    HISTORY_MAX_SIZE = 10

    def __init__(self, app, package, version, data_dir):
        Gtk.Window.__init__(self, application=app)
        self.package = package
        self.version = version
        self.data_dir = data_dir
        #setup the window
        self.set_default_size(600, 480)
        self.set_title(self.package)
        self.set_icon_name(package)

        signal_dict = {
            'action_systembus_connect_activate_cb': self.__systembus_connect_cb,
            'action_sessionbus_connect_activate_cb': self.__sessionbus_connect_cb,
            'action_otherbus_connect_activate_cb': self.__otherbus_connect_cb,
            'action_close_activate_cb': self.__close_cb,
            }

        #get settings
        settings = Settings.get_instance()

        ui = UILoader(self.data_dir, UILoader.UI_MAINWINDOW)
        vbox1 = ui.get_widget('vbox1')
        self.add(vbox1)
        self.connect('delete-event', self.__quit_dfeet)
        self.set_default_size(int(settings.general['windowwidth']),
                              int(settings.general['windowheight']))

        self.notebook = ui.get_widget('display_notebook')
        self.notebook.show_all()
        self.notebook_page_widget = ui.get_widget('box_notebook_page')
        #create bus history list and load entries from settings
        self.__bus_history = []
        for bus in settings.general['addbus_list']:
            if bus != '':
                self.__bus_history.append(bus)

        ui.connect_signals(signal_dict)
        #add a System- and Session Bus tab
        self.__systembus_connect_cb(None)
        self.__sessionbus_connect_cb(None)

        self.show_all()

    @property
    def bus_history(self):
        return self.__bus_history

    @bus_history.setter
    def bus_history(self, history_new):
        self.__bus_history = history_new

    def __systembus_connect_cb(self, action):
        """connect to system bus"""
        try:
            bw = BusWatch(self.data_dir, Gio.BusType.SYSTEM)
            self.__notebook_append_page(bw.paned_buswatch, "System Bus")
        except Exception as e:
            print(e)

    def __sessionbus_connect_cb(self, action):
        """connect to session bus"""
        try:
            bw = BusWatch(self.data_dir, Gio.BusType.SESSION)
            self.__notebook_append_page(bw.paned_buswatch, "Session Bus")
        except Exception as e:
            print(e)

    def __otherbus_connect_cb(self, action):
        """connect to other bus"""
        dialog = AddConnectionDialog(self.data_dir, self, self.bus_history)
        result = dialog.run()
        if result == Gtk.ResponseType.OK:
            address = dialog.address
            if address == 'Session Bus':
                self.__sessionbus_connect_cb(None)
                return
            elif address == 'System Bus':
                self.__systembus_connect_cb(None)
                return
            else:
                try:
                    bw = BusWatch(self.data_dir, address)
                    self.__notebook_append_page(bw.paned_buswatch, address)
                    # Fill history
                    if address in self.bus_history:
                        self.bus_history.remove(address)
                    self.bus_history.insert(0, address)
                    # Truncating history
                    if (len(self.bus_history) > self.HISTORY_MAX_SIZE):
                        self.bus_history = self.bus_history[0:self.HISTORY_MAX_SIZE]
                except Exception as e:
                    print("can not connect to '%s': %s" % (address, str(e)))
        dialog.destroy()

    def __notebook_append_page(self, widget, text):
        """add a page to the notebook"""
        ntl = NotebookTabLabel(text)
        page_nbr = self.notebook.append_page(widget, ntl)
        ntl.connect("close-clicked", self.__notebook_page_close_clicked_cb, widget)

    def __notebook_page_close_clicked_cb(self, button, widget):
        """remove a page from the notebook"""
        nbr = self.notebook.page_num(widget)
        self.notebook.remove_page(nbr)

    def __close_cb(self, action):
        """quit program"""
        self.__quit_dfeet(self, None)

    def __quit_dfeet(self, main_window, event):
        """quit d-feet application and store some settings"""
        settings = Settings.get_instance()
        size = main_window.get_size()
        pos = main_window.get_position()

        settings.general['windowwidth'] = size[0]
        settings.general['windowheight'] = size[1]

        self.bus_history = self.bus_history[0:self.HISTORY_MAX_SIZE]

        settings.general['addbus_list'] = self.bus_history
        settings.write()



#!/usr/bin/env python
import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(__file__, "../../")))

from gi.repository import Gtk
from gi.repository import Gio
from gi.repository import GLib
from dfeet.introspection import AddressInfo
from dfeet.introspection_helper import DBusNode
from dfeet.introspection_helper import DBusInterface
from dfeet.introspection_helper import DBusProperty
from dfeet.introspection_helper import DBusSignal
from dfeet.introspection_helper import DBusMethod
import unittest

XML = """
<node>
  <interface name='org.gnome.d-feet.TestInterface'>
    <method name='HelloWorld'>
      <arg type='s' name='greeting' direction='in'/>
      <arg type='s' name='response' direction='out'/>
    </method>
    <property name="TestString" type="s" access="readwrite"/>
    <property name="TestBool" type="b" access="read"/>
    <property name="TestStruct" type="(suu)" access="read"/>
    <signal name="TestSignal">
      <arg name="SigString" type="s"/>
      <arg name="SigBool" type="b"/>
    </signal>
  </interface>
</node>
"""

DATA_DIR = os.path.abspath("../../data/")


class IntrospectionHelperTest(unittest.TestCase):
    """tests for the introspection helper classes"""
    def setUp(self):
        self.name = "org.gnome.d-feet"
        self.object_path = "/org/gnome/d-feet"
        self.node_info = Gio.DBusNodeInfo.new_for_xml(XML)

    def test_dbus_node_info(self):
        """test DBusNode class"""
        obj_node = DBusNode(self.name, self.object_path, self.node_info)
        self.assertEqual(obj_node.name, self.name)
        self.assertEqual(obj_node.object_path, self.object_path)
        self.assertEqual(len(obj_node.node_info.interfaces), 1)

    def test_dbus_interface(self):
        """test DBusInterface class"""
        obj_node = DBusNode(self.name, self.object_path, self.node_info)
        obj_iface = DBusInterface(obj_node, obj_node.node_info.interfaces[0])
        self.assertEqual(obj_iface.name, self.name)
        self.assertEqual(obj_iface.object_path, self.object_path)
        self.assertEqual(len(obj_iface.iface_info.methods), 1)
        self.assertEqual(len(obj_iface.iface_info.properties), 3)
        self.assertEqual(len(obj_iface.iface_info.signals), 1)

    def test_dbus_property(self):
        """test DBusProperty class"""
        obj_node = DBusNode(self.name, self.object_path, self.node_info)
        obj_iface = DBusInterface(obj_node, obj_node.node_info.interfaces[0])
        obj_prop = DBusProperty(obj_iface, obj_iface.iface_info.properties[0])
        self.assertEqual(obj_prop.name, self.name)
        self.assertEqual(obj_prop.object_path, self.object_path)
        #get the markup string with value for struct prop (see bgo #702593)
        obj_prop = DBusProperty(obj_iface, obj_iface.iface_info.properties[2])
        obj_prop.value = ("string", 1, 2)
        self.assertIn("'string', 1, 2", obj_prop.markup_str)

    def test_dbus_signal(self):
        """test DBusSignal class"""
        obj_node = DBusNode(self.name, self.object_path, self.node_info)
        obj_iface = DBusInterface(obj_node, obj_node.node_info.interfaces[0])
        obj_sig = DBusSignal(obj_iface, obj_iface.iface_info.signals[0])
        self.assertEqual(obj_sig.name, self.name)
        self.assertEqual(obj_sig.object_path, self.object_path)


class AddressInfoTest(unittest.TestCase):
    """tests for the AddressInfo class and the introspection stuff"""

    def setUp(self):
        self.bus = Gio.TestDBus()
        self.bus.unset()
        self.bus.up()

    def tearDown(self):
        self.bus.stop()

    def test_bus(self):
        """introspect a name on the system bus"""
        ai = AddressInfo(DATA_DIR, self.bus.get_bus_address(), "org.freedesktop.DBus")

    @unittest.skip("TODO:peer to peer test not implemented")
    def test_peer_to_peer(self):
        """test a p2p connection"""
        #TODO: setup a gdbus server and test a peer to peer connection
        #(see http://developer.gnome.org/gio/unstable/GDBusServer.html#gdbus-peer-to-peer)
        pass

if __name__ == "__main__":
    #run tests
    unittest.main()
